"""
Event resources.

"""
from marshmallow import fields, Schema

from microcosm_flask.paging import PageSchema


class EventSchema(Schema):
    clock = fields.Integer(allow_none=True)
    createdTimestamp = fields.Float(
        attribute="created_timestamp",
        required=True,
        allow_none=True,
    )
    id = fields.UUID(
        required=True,
    )
    version = fields.Integer(
        required=True,
    )
    parentId = fields.UUID(
        attribute="parent_id",
        required=False,
        allow_none=True,
    )
    updatedTimestamp = fields.Float(
        attribute="created_timestamp",
        required=True,
        allow_none=True,
    )


class SearchEventSchema(PageSchema):
    clock = fields.Integer()
    min_clock = fields.Integer()
    max_clock = fields.Integer()
    parent_id = fields.UUID()
    sort_by_clock = fields.Boolean()
    version = fields.Integer()



"""
Event model.

"""
from microcosm_postgres.models import Model
from microcosm_postgres.types import EnumType, Serial
from sqlalchemy import (
    CheckConstraint,
    Column,
    FetchedValue,
    ForeignKey,
    Index,
    Integer,
)
from sqlalchemy.dialects.postgresql import ARRAY
from sqlalchemy_utils import UUIDType


# preserve the SQLAlchemy metaclass
MetaClass = type(Model)


class ColumnAlias:
    """
    Descriptor to reference a column by a well known alias.

    """
    def __init__(self, name):
        self.name = name

    def __get__(self, cls, owner):
        return getattr(cls or owner, self.name)

    def __set__(self, cls, value):
        return setattr(cls, self.name, value)


def default_state(context):
    return (context.current_parameters["event_type"],)


def join_event_types(event_types):
    return ",".join(
        "'{}'".format(event_type.name)
        for event_type in event_types
    )


class BaseEvent:

    def is_similar_to(self, other):
        """
        Are two events similar enough to activate upserting?

        """
        return all((
            self.event_type == other.event_type,
            self.parent_id == other.parent_id,
            self.container_id == other.container_id
        ))


class EventMeta(MetaClass):
    """
    Construct event models using a metaclass.

    SQLAlchemy's declarative base assumes that extending the base class (e.g. `Model`)
    declares a new table. As such, model subtypes are best defined with through mixins
    or metaclasses. In this case, a metaclass is more appropriate because we want some
    of the event attributes to be generated dynamically.

    """
    def __new__(cls, name, bases, dct):
        """
        Generate a new event type from the provided class.

        Requires that the class dictionary include the following declarations:

         -  `__container__` a reference to the container model class
         -  `__eventtype__` a reference to the event type enumeration
         -  `__tablename__` the usual SQLAlchemy table name

         Can optionally include the following declarations:

         -  `__unique_parent__` a flag indicating whether or not a unique parent constraint is created
         May be set to False in cases like having a unique parent for each version of the event
         If the flag is set to False, a similar unique constraint should be set on the event class

        """
        if any(type(base) is EventMeta for base in bases):
            return super(EventMeta, cls).__new__(cls, name, bases, dct)

        # add model to expected bases
        bases = bases + (BaseEvent, Model,)

        # declare event columns and indexes
        dct.update(cls.make_declarations(
            cls,
            container_name=dct["__container__"].__tablename__,
            event_type=dct["__eventtype__"],
            table_name=dct["__tablename__"],
            table_args=dct.get("__table_args__", ()),
            unique_parent=dct.get("__unique_parent__", True)
        ))

        return super(EventMeta, cls).__new__(cls, name, bases, dct)

    def make_declarations(cls, container_name, event_type, table_name, table_args, unique_parent):
        """
        Declare columns and indexes.

        An event assumes the following:

         -  Each event belongs to (and has a foreign key to) another "container" table.

            Typically, the container table has immutable rows that define the context of an entity
            and events track state changes to rows of this table.

         -  Each event has a well-defined type, derived from an enumeration.

         -  Each event has a well-defined state, consisting of one or more enumerated values.

         -  Each event has an integer version, starting from one.

            Not every event uses versions; those that do will generally have a uniqueness contraint
            on some event types per version of a container row.

         -  Each event has a nullable parent event id where a null value represents the first event
            in a version and subsequent event have a unique parent id to ensure semantic ordering.

         -  Each event has a non-nullable serial clock to ensure total ordering.

        """
        container_id = "{}.id".format(container_name)
        container_id_name = "{}_id".format(container_name)
        parent_id = "{}.id".format(table_name)

        return {
            # columns
            container_id_name: Column(UUIDType, ForeignKey(container_id), nullable=False),
            "event_type": Column(EnumType(event_type), nullable=False),
            "clock": Column(Serial, server_default=FetchedValue(), nullable=False, unique=True),
            "parent_id": Column(UUIDType, ForeignKey(parent_id), nullable=True, unique=unique_parent),
            "state": Column(ARRAY(EnumType(event_type)), nullable=False, default=default_state),
            "version": Column(Integer, default=1, nullable=False),

            # shortcuts
            "container_id": ColumnAlias(container_id_name),
            "container_id_name": container_id_name,

            # indexes and constraints
            "__table_args__": table_args + cls.make_table_args(cls, table_name, container_id_name, event_type),
        }

    def make_table_args(cls, table_name, container_id_name, event_type):
        """
        Generate the event table's `__table_args__` value.

        """
        return cls.make_indexes(
            cls,
            table_name,
            container_id_name,
        ) + cls.make_state_machine_constraints(
            cls,
            table_name,
            event_type,
        ) + cls.make_column_constraints(
            cls,
            table_name,
            event_type,
        )

    def make_indexes(cls, table_name, container_id_name):
        """
        Declare expected indexes.

        """
        return (
            # logical clock is unique and indexed
            Index(
                "{}_unique_logical_clock".format(table_name),
                container_id_name,
                "clock",
                unique=True,
            ),
            # NB: it's often but (not always) appropriate to have a unique index on the
            # combination of container id, event type, and version; for now this should
            # be added by the user.
        )

    def make_state_machine_constraints(cls, table_name, event_type):
        """
        Enforce that each state machine defines a proper linked list.

        """
        return (
            # events must have a parent unless they are initial and its the first version
            CheckConstraint(
                name="require_{}_parent_id".format(table_name),
                sqltext="parent_id IS NOT NULL OR (version = 1 AND event_type IN ({}))".format(
                    join_event_types(item for item in event_type if item.is_initial),
                ),
            ),
        )

    def make_column_constraints(cls, table_name, event_type):
        """
        Event tables are polymorphic and cannot enforce column-level nullability.

        Fortunately, check contraints can enforce non-nullability by event type.

        """
        return tuple(
            CheckConstraint(
                name="require_{}_{}".format(table_name, column_name),
                sqltext="{} IS NOT NULL OR event_type NOT IN ({})".format(
                    column_name,
                    join_event_types(item for item in event_type.requires(column_name))
                ),
            )
            for column_name in event_type.required_column_names()
        )



"""
Accumulation functions.

"""
from enum import Enum


def as_enum(value, event_type):
    if isinstance(value, Enum):
        return value
    return event_type.__class__[value]


def current():
    """
    Return the current event type.

    """
    return lambda state, event_type: [event_type]


def keep():
    """
    Keep the current state.

    """
    return lambda state, event_type: state


def alias(other_event_type):
    """
    Return another event type.

    """
    return lambda state, event_type: [as_enum(other_event_type, event_type)]


def addition(*other_event_types):
    """
    Add event types

    """
    def _addition(state, event_type):
        to_add = {
            as_enum(other_event_type, event_type)
            for other_event_type in other_event_types
        }
        return sorted(state | to_add)
    return _addition


def difference(*other_event_types):
    """
    Remove event types

    """
    def _difference(state, event_type):
        to_add = {
            as_enum(other_event_type, event_type)
            for other_event_type in other_event_types
        }
        return sorted(state - to_add)
    return _difference


def compose(*funcs):
    """
    Compose multiple accumulations.

    """
    def _compose(state, event_type):
        new_state = state
        for func in funcs:
            new_state = func(set(new_state), event_type)
        return new_state
    return _compose


def union():
    """
    Return the aggregated event type.

    """
    return lambda state, event_type: sorted(state | {event_type, })



"""
Factory for events.

Allows event creation logic to be decoupled from controllers.

"""
from inflection import camelize
from microcosm_flask.conventions.encoding import with_context
from microcosm_flask.naming import name_for
from microcosm_flask.operations import Operation
from microcosm_pubsub.conventions import created
from werkzeug.exceptions import UnprocessableEntity


class EventInfo:
    """
    Encapsulate information needed to create an event.

    """
    def __init__(self, ns, sns_producer, event_type, parent=None, version=None):
        self.ns = ns
        self.sns_producer = sns_producer
        self.event_type = event_type
        self.parent = parent
        self.version = version
        self.state = None
        self.event = None

    def publish_event(self, media_type, **kwargs):
        """
        Publish that an event occurred so that other services can react.

        """
        uri = self.ns.url_for(Operation.Retrieve, **kwargs)
        self.sns_producer.produce(
            media_type=media_type,
            uri=uri,
        )


class EventFactory:
    """
    Base class for creating an event.

    """
    def __init__(self, event_store, default_ns=None, identifier_key=None):
        self.event_store = event_store
        self.default_ns = default_ns
        self.identifier_key = identifier_key

    def create(self, ns, sns_producer, event_type, parent=None, version=None, **kwargs):
        """
        Create an event, validating the underlying state machine.

        """
        event_info = EventInfo(ns or self.default_ns, sns_producer, event_type, parent, version)
        self.validate_required_fields(event_info, **kwargs)
        self.validate_transition(event_info, **kwargs)
        if not event_info.parent:
            event_info.parent = self.event_store.retrieve_most_recent(**kwargs)
        self.create_transition(event_info, **kwargs)
        return event_info.event

    def create_transition(self, event_info, **kwargs):
        """
        Process an event state transition.

        This function allows chaining of state transitions by feeding the output
        of `create_event` with another event type.

        :raises: IllegalStateTransitionError

        """
        self.process_state_transition(event_info)
        self.create_event(event_info, **kwargs)

    def validate_required_fields(self, event_info, **kwargs):
        """
        Validate type-specific required fields.

        """
        missing_required_fields = [
            required_field
            for required_field in event_info.event_type.value.requires
            if kwargs.get(required_field) is None
        ]
        if missing_required_fields:
            raise with_context(
                UnprocessableEntity("Validation error"), [
                    {
                        "message": "Missing required field: '{}'".format(
                            camelize(required_field, uppercase_first_letter=False),
                        ),
                        "field": camelize(required_field, uppercase_first_letter=False),
                        "reasons": [
                            "Event type '{}' requires '{}'".format(
                                event_info.event_type.name,
                                camelize(required_field, uppercase_first_letter=False),
                            ),
                        ],
                    }
                    for required_field in missing_required_fields
                ],
            )

    def validate_transition(self, event_info, **kwargs):
        """
        Allows implementations of event source to define custom validation.

        """
        pass

    def process_state_transition(self, event_info):
        """
        Process a state transition.

        Determines whether a state transition is legal, what version to use, and what state
        to persist.

        """
        state = set(event_info.parent.state) if event_info.parent else set()
        event_info.event_type.validate_transition(state)
        event_info.state = event_info.event_type.accumulate_state(state)
        if event_info.version is not None:
            return
        parent_version = event_info.parent.version if event_info.parent else None
        event_info.version = event_info.event_type.next_version(parent_version)

    def create_event(self, event_info, **kwargs):
        """
        Create the event and publish that it was created.

        If the event has a parent id, uses an upsert to handle concurrent operations
        that produce the *same* event.

        """
        parent_id = None if event_info.parent is None else event_info.parent.id

        # NB: setting the id here so that it can easily be mocked in tests
        instance = self.event_store.model_class(
            id=self.event_store.new_object_id(),
            event_type=event_info.event_type,
            parent_id=parent_id,
            state=event_info.state,
            version=event_info.version,
            **kwargs
        )

        event_info.event = self.create_instance(event_info, instance)

        event_info.publish_event(
            media_type=self.make_media_type(event_info),
            **self.make_uri_kwargs(event_info)
        )

    def create_instance(self, event_info, instance):
        if event_info.parent is None:
            return self.event_store.create(instance)
        else:
            return self.event_store.upsert_on_index_elements(instance)

    def make_media_type(self, event_info):
        return created("{}.{}".format(
            name_for(self.event_store.model_class),
            event_info.event.event_type.name,
        ))

    def make_uri_kwargs(self, event_info):
        uri_kwargs = dict(_external=True)
        uri_kwargs[self.identifier_key] = event_info.event.id
        return uri_kwargs



"""
Transition mini-language.

"""


def normalize(value):
    """
    Normalize string values into id functions.

    """
    if callable(value):
        return value
    return event(value)


class Nothing:

    def __call__(self, cls, state):
        return not(state)

    def __bool__(self):
        return False

    __nonzero__ = __bool__


class AllOf:

    def __init__(self, *args):
        self.args = args

    def __call__(self, cls, state):
        return all(normalize(arg)(cls, state) for arg in self.args)

    def __bool__(self):
        return all(arg for arg in self.args)

    __nonzero__ = __bool__


class AnyOf:

    def __init__(self, *args):
        self.args = args

    def __call__(self, cls, state):
        return any(normalize(arg)(cls, state) for arg in self.args)

    def __bool__(self):
        return any(arg for arg in self.args)

    __nonzero__ = __bool__


class ButNot:

    def __init__(self, arg):
        self.arg = arg

    def __call__(self, cls, state):
        return not normalize(self.arg)(cls, state)

    def __bool__(self):
        return True

    __nonzero__ = __bool__


class Event:

    def __init__(self, name):
        self.name = name

    def __call__(self, cls, state):
        return cls[self.name] in state

    def __bool__(self):
        return True

    __nonzero__ = __bool__


def event(name):
    """
    Mini grammar to match a specific event type.

    """
    return Event(name)


def any_of(*args):
    """
    Mini grammar to match a list of event types.

    """
    return AnyOf(*args)


def all_of(*args):
    """
    Mini grammar to match a list of event types.

    """
    return AllOf(*args)


def but_not(arg):
    """
    Mini grammar to not match a specific event type.

    """
    return ButNot(arg)


def nothing():
    """
    Mini grammar that matches nothing.

    """
    return Nothing()



"""
Event store.

"""
from microcosm_postgres.store import Store
from sqlalchemy.dialects.postgresql import insert

from microcosm_eventsource.errors import ConcurrentStateConflictError


class EventStore(Store):
    """
    Event persistence operations.

    """
    def retrieve_most_recent(self, **kwargs):
        """
        Retrieve the most recent by container id and event type.

        """
        container_id = kwargs.pop(self.model_class.container_id_name)
        return self._retrieve_most_recent(
            self.model_class.container_id == container_id,
        )

    def retrieve_most_recent_by_event_type(self, event_type, **kwargs):
        """
        Retrieve the most recent by container id and event type.

        """
        container_id = kwargs.pop(self.model_class.container_id_name)
        return self._retrieve_most_recent(
            self.model_class.container_id == container_id,
            self.model_class.event_type == event_type,
        )

    def upsert_index_elements(self):
        """
        Can be overriden by implementations of event source to upsert based on other index elements

        Requires a unique constraint to exist on the index elements

        """
        return ["parent_id"]

    def upsert_on_index_elements(self, instance):
        """
        Upsert an event by index elements.

        Uses ON CONFLICT ... DO NOTHING to handle uniqueness constraint violations without
        invalidating the current transactions completely.

        Depends on an unique constraint on index elements to find the resulting entry.

        """
        with self.flushing():
            insert_statement = insert(self.model_class).values(
                instance._members(),
            )
            upsert_statement = insert_statement.on_conflict_do_nothing(
                index_elements=self.upsert_index_elements(),
            )
            self.session.execute(upsert_statement)

        most_recent = self._retrieve_most_recent(
            *[
                getattr(self.model_class, elem) == getattr(instance, elem)
                for elem in self.upsert_index_elements()
            ]
        )

        if not most_recent.is_similar_to(instance):
            raise ConcurrentStateConflictError()

        return most_recent

    def _filter(self,
                query,
                event_type=None,
                clock=None,
                min_clock=None,
                max_clock=None,
                parent_id=None,
                version=None,
                **kwargs):
        """
        Filter events by standard criteria.

        """
        container_id = kwargs.pop(self.model_class.container_id_name, None)
        if container_id is not None:
            query = query.filter(self.model_class.container_id == container_id)
        if event_type is not None:
            query = query.filter(self.model_class.event_type == event_type)
        if clock is not None:
            query = query.filter(self.model_class.clock == clock)
        if min_clock is not None:
            query = query.filter(self.model_class.clock >= min_clock)
        if max_clock is not None:
            query = query.filter(self.model_class.clock <= max_clock)
        if parent_id is not None:
            query = query.filter(self.model_class.parent_id == parent_id)
        if version is not None:
            query = query.filter(self.model_class.version == version)

        return super(EventStore, self)._filter(query, **kwargs)

    def _order_by(self, query, sort_by_clock=False, **kwargs):
        """
        Order events by logical clock.

        """
        if sort_by_clock:
            return query.order_by(
                self.model_class.clock.desc(),
            )
        return query.order_by(
            self.model_class.container_id.desc(),
            self.model_class.clock.desc(),
        )

    def _retrieve_most_recent(self, *criterion):
        """
        Retrieve the most recent event by some criterion.

        Note that the default `_order_by` enforces clock ordering.

        """
        return self._order_by(self._query(
            *criterion
        )).first()



"""
Event controllers.

"""
from microcosm_flask.conventions.crud_adapter import CRUDStoreAdapter

from microcosm_eventsource.factory import EventFactory


class EventController(CRUDStoreAdapter):

    @property
    def sns_producer(self):
        return self.graph.sns_producer

    @property
    def event_factory(self):
        """
        By default, create a new event factory.

        """
        return EventFactory(
            event_store=self.store,
            identifier_key=self.identifier_key,
        )

    def create(self, event_type, **kwargs):
        """
        Delegate to the event factory.

        """
        return self.event_factory.create(self.ns, self.sns_producer, event_type, **kwargs)



"""
Event type enumerations.

As events are processed, they accumulate a `state` as a list of event types, using
a pluggable rule system:

 -  The state can be the current event: `current()`
 -  The state can be a union of the current event and the current state: `union()`
 -  The state can be another state: `alias()`

State machine transitions are defined using a DNF-compatible mini-language that processes
this accumulated `state` and the current event:

 -  An event can legally follow another another event: `event(name)`
 -  An event can legally follow a disjunction of conditions: `any_of(...)`
 -  An event can legally follow a conjunction of conditions: `all_of(...)`
 -  An event can legally follow a negation of a conditon: `but_not(...)`

"""
from enum import Enum
from itertools import chain

from microcosm_eventsource.errors import (
    IllegalInitialStateError,
    IllegalStateTransitionError,
)
from microcosm_eventsource.accumulation import current
from microcosm_eventsource.transitioning import nothing


class EventTypeInfo:
    """
    Event type meta data.

    """
    def __init__(self, follows=None, accumulate=None, restarting=False, requires=()):
        """
        :param follows:    an instance of the event mini-grammar
        :param accumulate: whether the event type should accumulating state
        :param restarting: whether the event type may restart a (new) version

        """
        self.follows = follows or nothing()
        self.accumulate = accumulate or current()
        self.restarting = restarting
        self.requires = requires


class EventType(Enum):
    """
    Based event type enum.

    """
    @property
    def is_initial(self):
        """
        Can this event type be used for an initial event?

        """
        return not bool(self.value.follows)

    @property
    def is_accumulating(self):
        """
        Does this event type accumulate state?

        """
        return self.value.accumulating

    @property
    def is_restarting(self):
        """
        Does this event type restart a new version?

        """
        return self.value.restarting

    @classmethod
    def required_column_names(cls):
        return {
            column_name
            for event_type in cls
            for column_name in event_type.value.requires
        }

    @classmethod
    def requires(cls, column_name):
        return [
            event_type
            for event_type in cls
            if column_name in event_type.value.requires
        ]

    def may_transition(self, state):
        """
        Can this event type transition from the given state?

        :param state: a set of event types

        """
        return self.value.follows(self.__class__, state)

    def validate_transition(self, state):
        """
        Asssert that a transition is legal from the given state.

        :param state: a set of event types

        """
        if self.may_transition(state):
            return

        if not state:
            # event may not be initial
            raise IllegalInitialStateError("Event type '{}' may not be initial".format(
                self.name,
            ))

        # event may not follow previous
        raise IllegalStateTransitionError("Event type '{}' may not follow [{}]".format(
            self.name,
            ", ".join(event_type.name for event_type in state),
        ))

    def accumulate_state(self, state):
        """
        Accumulate state.

        """
        return self.value.accumulate(state, self)

    def next_version(self, version):
        """
        Compute next version.

        """
        if version is None:
            return 1
        if self.is_restarting:
            return version + 1
        return version

    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        return self.name


def event_info(*args, **kwargs):
    return EventTypeInfo(*args, **kwargs)


def EventTypeUnion(name, *event_types):
    """
    Create a new `EventType` as a union of other enums.

    """
    return EventType(name, [(item.name, item.value) for item in chain(*event_types)])



"""
Event routes

"""
from microcosm_flask.conventions.base import EndpointDefinition
from microcosm_flask.conventions.crud import configure_crud
from microcosm_flask.operations import Operation
from microcosm_postgres.context import transactional


def configure_event_crud(graph,
                         controller,
                         event_schema,
                         new_event_schema,
                         search_event_schema):
    mappings = {
        Operation.Create: EndpointDefinition(
            func=transactional(controller.create),
            request_schema=new_event_schema,
            response_schema=event_schema,
        ),
        Operation.Delete: EndpointDefinition(
            func=transactional(controller.delete),
        ),
        Operation.Replace: EndpointDefinition(
            func=transactional(controller.replace),
            request_schema=new_event_schema,
            response_schema=event_schema,
        ),
        Operation.Retrieve: EndpointDefinition(
            func=controller.retrieve,
            response_schema=event_schema,
        ),
        Operation.Search: EndpointDefinition(
            func=controller.search,
            request_schema=search_event_schema,
            response_schema=event_schema,
        ),
    }
    configure_crud(graph, controller.ns, mappings)
    return controller.ns



"""
State machine errors.

"""


class ConcurrentStateConflictError(Exception):
    @property
    def status_code(self):
        return 409


class IllegalStateTransitionError(Exception):
    @property
    def status_code(self):
        return 403


class IllegalInitialStateError(IllegalStateTransitionError):
    pass



#!/usr/bin/env python
from setuptools import find_packages, setup

project = "microcosm-eventsource"
version = "1.3.0"

setup(
    name=project,
    version=version,
    description="Event-sourced state machines using microcosm",
    author="Globality Engineering",
    author_email="engineering@globality.com",
    url="https://github.com/globality-corp/microcosm-eventsource",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    zip_safe=False,
    keywords="microcosm",
    install_requires=[
        "microcosm>=2.0.0",
        "microcosm-flask>=1.0.1",
        "microcosm-logging>=1.0.0",
        "microcosm-postgres>=1.0.0",
        "microcosm-pubsub>=1.0.0",
    ],
    setup_requires=[
        "nose>=1.3.6",
    ],
    dependency_links=[
    ],
    entry_points={
        "microcosm.factories": [
        ],
    },
    tests_require=[
        "coverage>=3.7.1",
        "mock>=2.0.0",
        "PyHamcrest>=1.9.0",
    ],
)
